# Copyright (C) 2018-2021 Shaon Ghosh, Shasvath Kapadia, Deep Chatterjee
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


import os
import pickle
import re

from argparse import ArgumentParser
from configparser import ConfigParser
import glob
import sqlite3

import numpy as np
import pandas as pd
from astropy.table import Column, Table, vstack
from sklearn.model_selection import KFold
from sklearn.neighbors import KNeighborsClassifier

from . import (
    EOS_MAX_MASS,
    EOS_BAYES_FACTORS,
    computeDiskMass
)


def join():
    """
    Joins the extracted sim-coinc data from GstLAL injection
    campaigns into a single Astropy table
    """
    parser = ArgumentParser(
        "Join extracted GstLAL sim-coinc parameters as astropy table")
    parser.add_argument("-i", "--input", required=True,
                        help="Directory storing extracted sim-coinc files")
    parser.add_argument("-c", "--config", required=True,
                        help="Name of the config file")
    parser.add_argument("-o", "--output", required=True,
                        help="Name of the output file")
    args = parser.parse_args()

    config = ConfigParser()
    config.read(args.config)
    extract_prefix = config.get('output_filenames',
                                'em_bright_extract_prefix')
    extract_suffix = config.get('output_filenames',
                                'em_bright_extract_suffix')
    cols = config.get('core',
                      'sqlite_cols')
    list_of_files = glob.glob(
        os.path.join(args.input, '{}*{}'.format(extract_prefix,
                                                extract_suffix)))
    cols = cols.split(',')
    # FIXME delimited '|' is a fragile piece
    data = vstack(
        [Table.read(f, format='ascii', delimiter='|', names=cols)
         for f in list_of_files]
    )
    # FIXME need a better implementation
    # convert injected detector frame mass to source frame masses
    inj_mass1_source = Column(data['inj_m1']/(1.0 + data['inj_redshift']),
                              name='inj_mass1_source_frame')
    inj_mass2_source = Column(data['inj_m2']/(1.0 + data['inj_redshift']),
                              name='inj_mass2_source_frame')
    ID = Column(np.arange(len(data)), name='id')
    data.add_column(ID, index=0)
    data.add_column(inj_mass1_source, index=0)
    data.add_column(inj_mass2_source, index=0)

    data.write(args.output, format='ascii', delimiter='\t')


def extract():
    """
    Ingests a GstLAL injection campaign sqlite database and
    outputs the list of coinc parameters in a flat file
    """
    parser = ArgumentParser(
        "Get sim-coinc maps for LIGO GstLAL injection sqlite database")
    parser.add_argument("-i", "--input", required=True,
                        help="sqlite database")
    parser.add_argument("-o", "--output", required=True,
                        help="Output file, stored as numpy array")
    args = parser.parse_args()

    cur = sqlite3.connect(args.input).cursor()
    cur.execute(
        """
        CREATE TEMPORARY TABLE
        sim_coinc_map_helper
        AS
        SELECT a.event_id as sid,
        coinc_event.coinc_event_id as cid,
        coinc_event.likelihood as lr
        FROM coinc_event_map as a
        JOIN coinc_event_map AS b ON (b.coinc_event_id == a.coinc_event_id)
        JOIN coinc_event ON (coinc_event.coinc_event_id == b.event_id)
        WHERE a.table_name == 'sim_inspiral'
        AND b.table_name == 'coinc_event'
        AND NOT EXISTS (
            SELECT * FROM time_slide WHERE
            time_slide.time_slide_id == coinc_event.time_slide_id
            AND time_slide.offset != 0
        )"""
    )

    cur.execute(
        """
        CREATE INDEX IF NOT EXISTS
        sim_coinc_map_helper_index ON sim_coinc_map_helper (sid, cid)
        """
    )

    cur.execute(
        """
        CREATE TEMPORARY TABLE
            sim_coinc_map
        AS
            SELECT
                sim_inspiral.simulation_id AS simulation_id,
                (
                    SELECT cid FROM
                                sim_coinc_map_helper
                    WHERE sid = simulation_id
                    ORDER BY lr DESC
                    LIMIT 1
                ) AS coinc_event_id
            FROM sim_inspiral
            WHERE coinc_event_id IS NOT NULL
        """
    )

    cur.execute("""DROP INDEX sim_coinc_map_helper_index""")

    query = """
    SELECT
    sim_inspiral.mass1,
    sim_inspiral.mass2,
    sim_inspiral.spin1z,
    sim_inspiral.spin2z,
        sim_inspiral.alpha3,
    sngl_inspiral.mass1,
    sngl_inspiral.mass2,
    sngl_inspiral.spin1z,
    sngl_inspiral.spin2z,
    sngl_inspiral.Gamma1,
    coinc_inspiral.combined_far,
    coinc_inspiral.snr,
    coinc_inspiral.end_time
    FROM
    sim_coinc_map
    JOIN
    sim_inspiral
    ON
    sim_coinc_map.simulation_id==sim_inspiral.simulation_id
    JOIN
    coinc_event_map
    ON
    sim_coinc_map.coinc_event_id == coinc_event_map.coinc_event_id
    JOIN
    coinc_inspiral
    ON
    sim_coinc_map.coinc_event_id == coinc_inspiral.coinc_event_id
    JOIN
    sngl_inspiral
    ON (
        coinc_event_map.table_name == 'sngl_inspiral'
        AND coinc_event_map.event_id == sngl_inspiral.event_id
    )
    WHERE
    sngl_inspiral.ifo=='H1';
    """
    np.savetxt(args.output, np.array(cur.execute(query).fetchall()),
               fmt='%f|%f|%f|%f|%f|%f|%f|%f|%f|%d|%e|%f|%f')


def train():
    parser = ArgumentParser(
        description='Executable to train source classifier from injections')

    parser.add_argument(
        '-i', '--input',
        help='Pickled dataframe containing source categorized data')
    parser.add_argument(
        '-o', '--output',
        help='Pickled object storing the trained classifiers')
    parser.add_argument(
        '-d', '--param-sweep-plot-prefix', default=None,
        help='Supply filename prefix to output a parameter sweep plot')
    parser.add_argument(
        '-c', '--config', required=True,
        help='Config file with additional parameters')

    args = parser.parse_args()

    config = ConfigParser()
    config.read(args.config)
    # compulsory sections in config
    required_sections = ['core',
                         'em_bright']
    assert all(config.has_section(s) for s in required_sections), \
        'Config file must have sections %s' % (required_sections,)

    # get column names and values from config
    feature_cols = config.get('em_bright',
                              'feature_cols').split(',')
    category_cols = config.get('em_bright',
                               'category_cols').split(',')
    threshold_cols = config.get('em_bright',
                                'threshold_cols').split(',')
    all_cols = feature_cols + category_cols + threshold_cols

    threshold_values = map(
        eval, config.get(
            'em_bright',
            'threshold_values').split(',')
        )
    threshold_type = config.get('em_bright',
                                'threshold_type').split(',')
    # read dataframe, check sanity
    with open(args.input, 'rb') as f:
        df = pickle.load(f)

    assert all(col in df.keys() for col in all_cols), \
        'Dataframe must contain columns %s' % (all_cols,)

    # create masked array based on threshold values,
    # extract features and targets
    mask = np.ones(len(df)).astype(bool)
    for col, value, typ in zip(threshold_cols, threshold_values,
                               threshold_type):
        mask &= df[col] < value if typ == 'lesser' else \
            df[col] > value if typ == 'greater' else True
    features = df[feature_cols][mask]
    targets = df[category_cols][mask]
    # train Nearest Neighbor classifier for each category
    clf_kwargs = eval(config.get('em_bright',
                                 'clf_kwargs'))
    if clf_kwargs.get('metric') == 'mahalanobis':
        clf_kwargs['metric_params'] = dict(
            V=features.cov().values
        )  # covariance matrix needed for mahalanobis metric
    clfs = []
    for category, target_value in targets.iteritems():
        # run KFold cross-validation
        res_predict, res_predict_proba = run_k_fold_split(features,
                                                          target_value,
                                                          **clf_kwargs)
        test_results = pd.DataFrame(
            np.vstack((features.T, target_value,
                       res_predict, res_predict_proba)).T,
            columns=feature_cols + ['targets', 'predict', 'predict_proba']
        )
        test_results_filename = 'cross-val-res-' + category + '.csv'
        test_results.to_csv(test_results_filename, index=False)
        # train on the full dataset
        clf = KNeighborsClassifier(**clf_kwargs)
        clf.fit(features, target_value)
        if args.param_sweep_plot_prefix:
            _create_param_sweep_plot(
                clf, category,
                args.param_sweep_plot_prefix
            )
        clfs.append(clf)
    # append the output filename of the classifier
    clfs.extend([args.output])
    with open(args.output, 'wb') as f:
        pickle.dump(clfs, f)


def _open_and_return_clfs(filename):
    """Unpack pickle files storing classifier and return.
    First return argument is HasNS, second HasRemnant
    """
    with open(filename, 'rb') as f:
        clf_ns, clf_em, _filename = pickle.load(f)
    return clf_ns, clf_em


def _get_mass_grid():
    """Get a grid over mass1, mass2. Used for parameter sweep."""
    mass1 = np.linspace(1, 20, 200)
    mass2 = np.linspace(1, 20, 200)
    t = Table(
        data=np.vstack(
            (np.repeat(mass1, mass2.size),
             np.tile(mass2, mass1.size))
        ).T, names=('mass1', 'mass2')
    )
    mask = t['mass1'] > t['mass2']
    t = t[mask]
    return t


def _get_param_sweep(clf):
    """Create a fake recovered parameter space return
    the predictions for the classifier sweeping across
    masses.
    """
    if clf.metric == "mahalanobis":
        clf.n_jobs = 1  # issue with BallTree output
    t = _get_mass_grid()
    spins = Table(
        data=np.vstack(
            (np.repeat(np.linspace(0, 1, 2), 2),
             np.tile(np.linspace(0, 1, 2), 2))
        ).T,
        names=('chi1', 'chi2')
    )
    SNR = 10.
    res = list()
    for spin_vals in spins:
        SNR *= np.ones(len(t))
        chi1 = spin_vals['chi1'] * np.ones(len(t))
        chi2 = spin_vals['chi2'] * np.ones(len(t))
        # make predictions and make plots
        param_sweep_features = np.stack(
            [t['mass1'], t['mass2'], chi1, chi2, SNR]
        ).T
        predictions = clf.predict_proba(param_sweep_features).T[1]
        res.append((param_sweep_features, predictions))
    return res


def _create_param_sweep_plot(clf, category, prefix=None):
    """Create a parameter sweep plot using the supplied classifer"""
    import matplotlib.pyplot as plt
    res = _get_param_sweep(clf)
    fig = plt.figure(figsize=(14, 20))
    for idx, r in enumerate(res):
        features, predictions = r
        # FIXME: Ugly, but works
        title = "chi1 = {0}; chi2 = {1}; SNR = {2}".format(
            features[0][2], features[0][3],
            features[0][4]
        )
        make_plots(
            features, predictions, title,
            (fig, idx+1), prefix=prefix, category=category
        )
    try:
        plt.savefig(prefix+'_param_sweep_'+category+'.png')
    except TypeError:
        plt.savefig('param_sweep_'+category+'.png')


def param_sweep_plot():
    """Create parameter sweep plot, weighting classifier
    results using bayes factor
    """
    parser = ArgumentParser(
        "Create a parameter sweep, EoS predictions based")
    parser.add_argument("-i", "--input", required=True,
                        help="Directory storing trained classifier files")
    parser.add_argument("-c", "--config", required=True,
                        help="Config file")
    parser.add_argument("-v", "--verbose", action='store_true',
                        help="show progress")
    args = parser.parse_args()

    config = ConfigParser()
    config.read(args.config)

    train_prefix = config.get('output_filenames', 'em_bright_train_prefix')
    train_suffix = config.get('output_filenames', 'em_bright_train_suffix')
    clf_filenames = glob.glob(
        os.path.join(args.input, f'{train_prefix}*{train_suffix}'))
    if args.verbose:
        print("Trained classifers", clf_filenames)
    assert clf_filenames, "No files found. Check directory."
    reweight_ns = dict.fromkeys(EOS_BAYES_FACTORS)
    reweight_em = dict.fromkeys(EOS_BAYES_FACTORS)
    import matplotlib.pyplot as plt
    for fname in clf_filenames:
        if args.verbose:
            print(fname)
        clf_ns, clf_em = _open_and_return_clfs(fname)
        eosname, *_ = filter(
            lambda _: re.match(
                ".*" + _ + config.get('output_filenames',
                                      'em_bright_train_suffix'),
                fname
            ), EOS_BAYES_FACTORS
        )
        res_ns = _get_param_sweep(clf_ns)
        res_em = _get_param_sweep(clf_em)
        reweight_ns.update(
            {
                eosname: dict(
                    features=np.array(
                        [f for f, p in res_ns]
                    ),
                    predictions=np.array(
                        [p * EOS_BAYES_FACTORS[eosname] for f, p in res_ns]
                    )
                )
            }
        )
        reweight_em.update(
            {
                eosname: dict(
                    features=np.array(
                        [f for f, p in res_em]
                    ),
                    predictions=np.array(
                        [p * EOS_BAYES_FACTORS[eosname] for f, p in res_em]
                    )
                )
            }
        )
    assert all(reweight_ns.values()), "Missing some listed EOSs."
    assert all(reweight_em.values()), "Missing some listed EOSs."
    reweighted_score_ns = np.sum(
        [v['predictions'] for v in reweight_ns.values()],
        axis=0
    )
    reweighted_score_em = np.sum(
        [v['predictions'] for v in reweight_em.values()],
        axis=0
    )
    for category, score in zip(
            ["NS", "EMB"], [reweighted_score_ns, reweighted_score_em]):
        fig = plt.figure(figsize=(14, 20))
        for idx, (f, p) in enumerate(
                zip(reweight_ns['SLy']['features'], score)):
            title = "chi1 = {0}; chi2 = {1}; SNR = {2}".format(
                f[0][2], f[0][3], f[0][4])
            make_plots(
                f, p, title, (fig, idx+1),
                prefix='SLy', category=category
            )
        plt.savefig("reweighted_" + category + "_param_sweep.png")


def make_plots(features, predictions, title, fig_idx,
               prefix=None, category=None):
    import matplotlib.pyplot as plt
    fig_, idx = fig_idx
    fig_.add_subplot(4, 1, idx)
    # indices 0 and 1 correspond to mass1 and mass2 respectively
    plt.scatter(features.T[0], features.T[1],
                s=10, c=predictions)
    plt.title(title)
    plt.tight_layout()
    plt.colorbar(label='Probability')
    # plot chirp mass contours
    m1 = np.linspace(1, 20, 1000)
    m2 = np.linspace(1, 20, 1000)
    M1, M2 = np.meshgrid(m1, m2)
    s1z = np.unique(features.T[2])[0]*np.ones(M1.shape)
    s2z = np.unique(features.T[3])[0]*np.ones(M2.shape)
    rem_masses = list()
    for _ in range(M1.shape[0]):
        rem_masses.append(
            computeDiskMass.computeDiskMass(
                M1[_], M2[_], s1z[_], s2z[_],
                eosname=prefix
            )
        )
    rem_masses = np.array(rem_masses).reshape(M1.shape)
    Mc = (M1*M2)**(3./5.)/(M1 + M2)**(1./5.)
    mask = M1 > M2
    M1 = np.ma.masked_array(M1, mask=mask)
    M2 = np.ma.masked_array(M2, mask=mask)
    s1z = np.ma.masked_array(s1z, mask=mask)
    s2z = np.ma.masked_array(s2z, mask=mask)
    Mc = np.ma.masked_array(Mc, mask=mask)
    rem_masses = np.ma.masked_array(
        rem_masses, mask=mask*(M2 > EOS_MAX_MASS[prefix])
    )
    CS = plt.contour(
        M1, M2, Mc.T, levels=[2.22, 2.99, 4.73, 5, 6],
        # (m1, m2) = (5, 1.4) -> Mc = 2.22
        # (m1, m2) = (10, 1.4) -> Mc = 2.99
        # (m1, m2) = (30, 1.4) -> Mc = 4.73
        # three different NSBH populations; remaining values are ad-hoc
        colors='black', linewidths=1.0
    )
    plt.clabel(CS, inline=True, fontsize=16)
    plt.xlim((1, 20))
    plt.ylim((1, 12))
    try:
        max_mass = EOS_MAX_MASS[prefix]
    except KeyError:
        max_mass = 3.0
    if 'NS' in category:
        plt.axhline(y=max_mass, c='r', linewidth=1.2)
    elif 'EM' in category:
        CS = plt.contour(M1, M2, rem_masses.T,
                         levels=[0., ], colors='red',
                         linewidths=1.2)
    plt.xlabel(r'$m_1$', fontsize=16)
    plt.ylabel(r'$m_2$', fontsize=16)


def run_k_fold_split(features, targets, n_splits=10, random_state=0,
                     **kwargs):
    """Split the `features` in `n_splits`, train on `n_splits - 1`
    sets, test on the last fraction. This performed across the complete
    dataset.

    Parameters
    ----------
    features : numpy.ndarray
        Feature set
    targets : numpy.ndarray
        Target set containing binary classification of `features`
    n_splits : int
        Number of splits for `features` and `targets`
    random_state : int
        Random seef for the split
    **kwargs
        Keyword arguments passed to `KNearestNeighborClassifier`

    Returns
    -------
    tuple
        (res_predict, res_predict_proba) - predictions and probabilities
    """
    res_predict_proba = np.empty(len(features))
    res_predict = np.empty(len(features))
    res_predict_proba[:] = np.nan
    res_predict[:] = np.nan

    sss = KFold(n_splits=n_splits, shuffle=True, random_state=random_state)
    sss.get_n_splits(features, targets)
    for train_index, test_index in sss.split(features, targets):
        X_train, X_test, y_train = \
            features.iloc[train_index], \
            features.iloc[test_index], \
            targets.iloc[train_index]

        predict_proba, predict = run_KNN_classifier(X_train, y_train, X_test,
                                                    **kwargs)
        # second column is the prob of NS/EMB
        predict_proba = predict_proba.T[1]

        res_predict_proba[test_index] = predict_proba
        res_predict[test_index] = predict
    return res_predict, res_predict_proba


def run_KNN_classifier(X_train, y_train,
                       X_test,
                       **kwargs):
    '''
    Run KNearestNeighborClassifier classifier returns
    `clf.predict_proba`

    Parameters
    ----------
    X_train : numpy.ndarray
        Feature training set, can be array or Dataframe
    y_train : numpy.array
        Target training set, 0 or 1 binary classification
    X_test : numpy.ndarray
        Feature testing set, can be array or DataFrame
    '''
    if kwargs.get('metric') == 'mahalanobis':
        kwargs['n_jobs'] = 1  # issue with KDTree, BallTree with n_jobs > 1
    clf = KNeighborsClassifier(**kwargs)
    clf.fit(X_train, y_train)
    predictions_proba = clf.predict_proba(X_test)
    predict = clf.predict(X_test)
    return predictions_proba, predict



from pathlib import Path
from tempfile import NamedTemporaryFile

import numpy as np
import h5py
import pandas as pd

import pytest
from unittest.mock import Mock

from .. import em_bright, categorize, EOS_MAX_MASS


def test_version():
    from .. import __version__
    assert __version__ == '1.0.2'


@pytest.mark.parametrize(
    'posteriors, dtype, result',
    [[[(1.2, 1.0, 0.0, 0.0, 0.0, 0.0, 100.0),
       (2.0, 0.5, 0.99, 0.99, 0.0, 0.0, 150.0)],
      [('mc', '<f8'), ('q', '<f8'), ('a1', '<f8'),
       ('a2', '<f8'), ('tilt1', '<f8'), ('tilt2', '<f8'),
       ('dist', '<f8')],
      (1.0, 1.0)],
     [[(1.2, 1.0, 0.0, 0.0, 100.0),
       (2.0, 0.5, 0.99, 0.99, 150.0)],
      [('mc', '<f8'), ('q', '<f8'), ('a1', '<f8'),
       ('a2', '<f8'), ('dist', '<f8')],
      (1.0, 1.0)]]
)
def test_source_classification_pe(posteriors, dtype, result):
    """Test em_bright classification from posterior
    samples - both aligned and precessing cases.
    """
    with NamedTemporaryFile() as f:
        filename = f.name
        with h5py.File(f, 'w') as tmp_h5:
            data = np.array(
                posteriors,
                dtype=dtype
            )
            tmp_h5.create_dataset(
                'lalinference/lalinference_mcmc/posterior_samples',
                data=data
            )
        r = em_bright.source_classification_pe(filename)
    assert r == result


@pytest.mark.parametrize(
    'm1,m2,chi1,chi2,snr,result_ns, result_em',
    [[1.1, 1.0, 0.0, 0.0, 20.0, 1.0, 1.0],
     [1.1, 1.0, 0.0, 0.0, 5.0, 1.0, 1.0],
     [100.0, 50.0, 0.0, 0.0, 10.0, 0.0, 0.0],
     [8.0, 1.4, -0.99, 0.0, 20.0, 1.0, 0.0],
     [8.0, 1.4, 0.99, 0.0, 20.0, 1.0, 1.0]]
)
def test_source_classification_sanity(m1, m2, chi1, chi2, snr,
                                      result_ns, result_em):
    '''Test results for limiting cases'''
    ns, em = em_bright.source_classification(m1, m2, chi1, chi2, snr)
    assert ns == pytest.approx(result_ns)
    assert em == pytest.approx(result_em)


def test_mock_classifier():
    '''Test to check usage of custom classifier'''
    m1 = 1.1
    m2 = 1.0
    chi1 = 0.04
    chi2 = 0.0
    snr = 20.
    # Define mock objects
    mock_clf_ns = Mock()
    mock_clf_emb = Mock()
    mock_clf_ns.predict_proba = Mock(return_value=np.array([[1., 1.]]))
    mock_clf_emb.predict_proba = Mock(return_value=np.array([[1., 1.]]))

    ns, em = em_bright.source_classification(m1, m2, chi1, chi2, snr,
                                             ns_classifier=mock_clf_ns,
                                             emb_classifier=mock_clf_emb)
    mock_clf_ns.predict_proba.assert_called_once()
    mock_clf_emb.predict_proba.assert_called_once()
    assert ns == 1.0
    assert em == 1.0


@pytest.mark.parametrize(
    'm1,m2,chi1,chi2,eosname,non_zero_remnant',
    [[1.1, 1.0, 0.0, 0.0, "2H", 1.0],
     [1.1, 1.0, 0.0, 0.0, "SLy", 1.0],
     [1.1, 1.0, 0.0, 0.0, "AP4", 1.0],
     [1.1, 1.0, 0.0, 0.0, "WFF1", 1.0],
     [26.0, 2.6, 0.0, 0.0, "2H", 0.0],
     [10.0, 1.4, 0.99, 0.0, "2H", 1.0],
     [10.0, 1.4, -0.99, 0.0, "2H", 0.0]]
)
def test_compute_disk_mass(m1, m2, chi1, chi2,
                           eosname, non_zero_remnant):
    has_remnant = em_bright.computeDiskMass.computeDiskMass(
        m1, m2, chi1, chi2, eosname=eosname
    ) > 0.
    assert has_remnant == non_zero_remnant


def test_embright_categorization():
    for eosname, max_mass in EOS_MAX_MASS.items():
        with NamedTemporaryFile() as tf:
            outFile = tf.name
        categorize.embright_categorization(
            Path(__file__).parents[0] / 'data/test_categorize_data.tbl',
            outFile, eosname=eosname
        )
        # read in pickle file
        res = pd.read_pickle(outFile)
        assert ~np.all(
            np.logical_xor(
                res.NS.values.astype(bool),
                res.m2_inj.values < EOS_MAX_MASS[eosname]
            )
        )



import json

from astropy.utils.data import download_file

__version__ = '1.0.2'

PACKAGE_DATA_LINKS = {
    'equil_2H.dat': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/equil_2H.dat',  # noqa: E501
    'APR4_EPP.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/APR4_EPP.pickle',  # noqa: E501
    'BHF_BBB2.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/BHF_BBB2.pickle',  # noqa: E501
    'H4.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/H4.pickle',  # noqa: E501
    'HQC18.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/HQC18.pickle',  # noqa: E501
    'KDE0V1.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/KDE0V1.pickle',  # noqa: E501
    'KDE0V.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/KDE0V.pickle',  # noqa: E501
    'MPA1.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/MPA1.pickle',  # noqa: E501
    'MS1B_PP.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/MS1B_PP.pickle',  # noqa: E501
    'MS1_PP.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/MS1_PP.pickle',  # noqa: E501
    'RS.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/RS.pickle',  # noqa: E501
    'SK255.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SK255.pickle',  # noqa: E501
    'SK272.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SK272.pickle',  # noqa: E501
    'SKI2.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SKI2.pickle',  # noqa: E501
    'SKI3.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SKI3.pickle',  # noqa: E501
    'SKI4.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SKI4.pickle',  # noqa: E501
    'SKI5.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SKI5.pickle',  # noqa: E501
    'SKI6.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SKI6.pickle',  # noqa: E501
    'SKMP.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SKMP.pickle',  # noqa: E501
    'SKOP.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SKOP.pickle',  # noqa: E501
    'SLY230A.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SLY230A.pickle',  # noqa: E501
    'SLY2.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SLY2.pickle',  # noqa: E501
    'SLY9.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SLY9.pickle',  # noqa: E501
    'SLy.pickle': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/SLy.pickle',  # noqa: E501
    'EOS_BAYES_FACTOR_MAP.json': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/EOS_BAYES_FACTOR_MAP.json',  # noqa: E501
    'EOS_MAX_MASS_MAP.json': 'https://git.ligo.org/deep.chatterjee/em-bright/-/raw/v1.0.2/ligo/em_bright/data/EOS_MAX_MASS_MAP.json'  # noqa: E501
}

PACKAGE_FILENAMES = dict.fromkeys(PACKAGE_DATA_LINKS)

for name, url in PACKAGE_DATA_LINKS.items():
    PACKAGE_FILENAMES[name] = download_file(
        url, cache=True, pkgname='ligo.em_bright'
    )

# load and normalize bayes factors
with open(PACKAGE_FILENAMES['EOS_BAYES_FACTOR_MAP.json']) as f:
    EOS_BAYES_FACTORS = json.load(f)
    """Taken from Table II (broad prior), Approx bayes factor column
    of 10.1103/PhysRevD.104.083003
    """

_BAYES_FACTOR_NORM = sum(EOS_BAYES_FACTORS.values())
"""Sum of bayes factors above"""

EOS_BAYES_FACTORS.update(
    {k: v / _BAYES_FACTOR_NORM for k, v in EOS_BAYES_FACTORS.items()}
)

with open(PACKAGE_FILENAMES['EOS_MAX_MASS_MAP.json']) as f:
    EOS_MAX_MASS = json.load(f)
    """Taken from Table I m_max column of 10.1103/PhysRevD.104.083003"""



# Copyright (C) 2020 Shaon Ghosh, Deep Chatterjee
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


import argparse
import functools
import os

import numpy as np
import pandas as pd
import pickle

from . import computeDiskMass, EOS_BAYES_FACTORS, EOS_MAX_MASS


class _TupleHandler(object):
    def __call__(self, func):
        @functools.wraps(func)
        def handle(*args, **kwargs):
            try:
                (infile, outfile, max_mass, eosname), *_ = args
            except (ValueError, TypeError):
                return func(*args, **kwargs)
            return func(infile, outfile, max_mass, eosname)
        return handle


def regularize(m1, m2, chi1, chi2):
    '''Transform masses based on the convention m1 > m2.
    Apply the same for spins.
    '''
    # Find cases where the mass2 is greater than the mass1
    largerMass1 = m1 >= m2
    largerMass2 = m2 > m1

    # Set up vectors to store the regularized masses and spins
    m_primary = np.zeros_like(m1)
    m_secondary = np.zeros_like(m2)
    chi_primary = np.zeros_like(chi1)
    chi_secondary = np.zeros_like(chi2)

    # Swap vales of masses if mass1 < mass2
    m_primary[largerMass1] = m1[largerMass1]
    m_primary[~largerMass1] = m2[largerMass2]
    m_secondary[~largerMass2] = m2[~largerMass2]
    m_secondary[largerMass2] = m1[~largerMass1]

    # Assign the spin values accordingly
    chi_primary[largerMass1] = chi1[largerMass1]
    chi_primary[~largerMass1] = chi2[~largerMass1]
    chi_secondary[~largerMass2] = chi2[~largerMass2]
    chi_secondary[largerMass2] = chi1[~largerMass1]

    return (m_primary, m_secondary, chi_primary, chi_secondary)


@_TupleHandler()
def embright_categorization(inFile, outFile, eosname='2H', mNs_mass=None):
    '''Categorize whether the binary has a neutron star, and any non-zero
    remnant post merger, based on the ``eosname``.

    Parameters
    ----------
    inFile : str
        input filename
    outFile : str
        output filename
    eosname : str
        neutron star equation of state. Assumed implemented in lalsimulation.
        Default 2H.
    mNS_mass : float
        Provide to override the maximum mass from ``eosname``.
    '''
    df = pd.read_table(inFile, delimiter='\t')
    m1_inj, m1_rec = df.inj_m1.values, df.rec_m1.values
    m2_inj, m2_rec = df.inj_m2.values, df.rec_m2.values
    redshift_inj = df.inj_redshift.values
    m1_inj_source = m1_inj / (1 + redshift_inj)
    m2_inj_source = m2_inj / (1 + redshift_inj)

    chi1_inj, chi1_rec = df.inj_spin1z.values, df.rec_spin1z.values
    chi2_inj, chi2_rec = df.inj_spin2z.values, df.rec_spin2z.values
    m1_inj, m2_inj, chi1_inj, chi2_inj = regularize(
        m1_inj, m2_inj, chi1_inj, chi2_inj
    )
    m1_rec, m2_rec, chi1_rec, chi2_rec = regularize(
        m1_rec, m2_rec, chi1_rec, chi2_rec
    )
    m1_inj_source, m2_inj_source, _, _ = regularize(
        m1_inj_source, m2_inj_source, chi1_inj, chi2_inj
    )  # spins regularized from previous step.
    if mNs_mass:
        NS_classified = m2_inj_source < mNs_mass
    else:
        NS_classified = m2_inj_source < EOS_MAX_MASS[eosname]
    NS_classified = NS_classified.astype(int)

    R_isco_hat_inj = computeDiskMass.compute_isco(chi1_inj)
    R_isco_hat_rec = computeDiskMass.compute_isco(chi1_rec)

    Compactness_inj, *_ = computeDiskMass.computeCompactness(m2_inj,
                                                             eosname=eosname)
    Compactness_rec, *_ = computeDiskMass.computeCompactness(m2_rec,
                                                             eosname=eosname)

    mc_inj = (m1_inj*m2_inj)**(3/5)/((m1_inj + m2_inj)**(1/5))
    mc_rec = (m1_rec*m2_rec)**(3/5)/((m1_rec + m2_rec)**(1/5))

    frac_mc_err = np.abs(mc_inj - mc_rec)/mc_inj

    q_inj = m1_inj/m2_inj
    q_rec = m1_rec/m2_rec

    disk_mass_inj = computeDiskMass.computeDiskMass(
        m1_inj_source, m2_inj_source, chi1_inj,
        chi2_inj, eosname=eosname
    )
    EMB_classified = disk_mass_inj > 1e-3
    EMB_classified = EMB_classified.astype(int)

    output = np.vstack(
        (m1_inj, m2_inj, chi1_inj, chi2_inj, redshift_inj,
         mc_inj, q_inj, R_isco_hat_inj, Compactness_inj,
         m1_rec, m2_rec, chi1_rec, chi2_rec,
         mc_rec, frac_mc_err, q_rec, R_isco_hat_rec, Compactness_rec,
         df.cfar.values, df.snr.values, df.gpstime.values, NS_classified,
         EMB_classified)
    ).T

    df_complete = pd.DataFrame(output, columns=[
        'm1_inj', 'm2_inj', 'chi1_inj', 'chi2_inj', 'inj_redshift',
        'mc_inj', 'q_inj', 'R_isco_inj', 'Compactness_inj',
        'm1_rec', 'm2_rec', 'chi1_rec', 'chi2_rec', 'mc_rec',
        'frac_mc_err', 'q_rec', 'R_isco_rec', 'Compactness_rec',
        'cfar', 'snr', 'gpstime', 'NS', 'EMB']
    )
    # drop NSs above max-mass
    try:
        max_mass = EOS_MAX_MASS[eosname]
    except KeyError:
        raise NotImplementedError
    drop_mask = (m1_inj_source > max_mass) & (m1_inj_source < 3.0)  # noqa: E501; Restricts m2 by construction
    df_complete = df_complete.loc[~drop_mask]
    with open(outFile, 'wb') as f:
        pickle.dump(df_complete, f)
    return df_complete


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", action="store", type=str,
                        help="Name of the input file")
    parser.add_argument("-o", "--output", action="store", type=str,
                        help="Name of the output file")
    parser.add_argument(
        "-e", "--eosname", default='2H',
        help="Equation of state used compute remnant matter."
    )
    args = parser.parse_args()

    embright_categorization(args.input, args.output,
                            eosname=args.eosname)


def main_all():
    parser = argparse.ArgumentParser(
        "Run categorization on all available EoSs: "
        f"{EOS_BAYES_FACTORS.keys()}\n"
        "Output will be tagged by EoS names."
    )
    parser.add_argument(
        "-i", "--input", action="store", type=str,
        help="Name of the input file"
    )
    parser.add_argument(
        "-d", "--output-directory", type=str,
        help="Output directory"
    )
    parser.add_argument(
        "-o", "--output-prefix", action="store",
        type=str, help="Input prefix "
    )
    parser.add_argument(
        "-p", "--pool", default=1, type=int,
        help="Pool size"
    )
    args = parser.parse_args()

    from multiprocessing import Pool
    eos_args = [
        (
            args.input, os.path.join(
                args.output_directory,
                f"{args.output_prefix}_{eosname}.pkl"
            ),
            EOS_MAX_MASS[eosname],
            eosname
        )
        for eosname in EOS_BAYES_FACTORS
    ]
    with Pool(args.pool) as p:
        p.map(embright_categorization, eos_args)



# Copyright (C) 2018-2022 Shaon Ghosh, Deep Chatterjee
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


import os
import re
import shutil

from argparse import ArgumentParser
from configparser import ConfigParser
from glob import glob

from . import EOS_BAYES_FACTORS

try:
    import htcondor
    from htcondor import dags
    _HTCONDOR_INSTALLED = True
except ModuleNotFoundError:
    _HTCONDOR_INSTALLED = False
    import warnings
    warnings.warn(
        "HTCondor python bindings need to be installed to use this script. "
        "This script is only used in training. When training new classifiers, "
        "create a developmental version by cloning the git repository, "
        "and installing from the lock file."
    )


def main():
    parser = ArgumentParser("Condor DAG writer for workflow")
    parser.add_argument("-d", "--dagname", required=True,
                        help="Name of the dag file. Placed under --work-dir")
    parser.add_argument("-w", "--work-dir", required=True,
                        help="Working directory to store data outputs")
    parser.add_argument(
        "-i", "--file-dir", required=True,
        help="File containing input injection sqlite databases"
    )
    parser.add_argument("-c", "--config", required=True,
                        help="Name of the config file")
    parser.add_argument("-e", "--executables-dir", required=True,
                        help="Directory containing executables")
    args = parser.parse_args()

    assert _HTCONDOR_INSTALLED, "HTCondor python bindings missing."

    config = ConfigParser()
    config.read(args.config)
    abs_config_file = os.path.abspath(args.config)

    # Get path for the work directory
    if not os.path.exists(args.work_dir):
        os.makedirs(args.work_dir)
    abs_work_dir = os.path.abspath(args.work_dir)

    # Get injection directory
    inj_file_pattern = config.get('core',
                                  'inj_file_pattern')
    sqlite_run_tag = config.get('core',
                                'sqlite_run_tag')
    accounting_group = config.get('core',
                                  'accounting_group')

    # Get input and output file names prefixes
    em_bright_extract_prefix = config.get('output_filenames',
                                          'em_bright_extract_prefix')
    em_bright_extract_suffix = config.get('output_filenames',
                                          'em_bright_extract_suffix')
    em_bright_join_prefix = config.get('output_filenames',
                                       'em_bright_join_prefix')
    em_bright_join_suffix = config.get('output_filenames',
                                       'em_bright_join_suffix')
    em_bright_categorize_prefix = config.get('output_filenames',
                                             'em_bright_categorize_prefix')
    em_bright_categorize_suffix = config.get('output_filenames',
                                             'em_bright_categorize_suffix')
    em_bright_train_prefix = config.get('output_filenames',
                                        'em_bright_train_prefix')
    em_bright_train_suffix = config.get('output_filenames',
                                        'em_bright_train_suffix')
    # Get executable names
    em_bright_extract_executable = config.get('executables',
                                              'em_bright_extract')
    em_bright_join_executable = config.get('executables',
                                           'em_bright_join')
    em_bright_categorize_executable = config.get('executables',
                                                 'em_bright_categorize')
    em_bright_train_executable = config.get('executables',
                                            'em_bright_train')
    em_bright_paramater_sweep_plot_executable = config.get(
        'executables', 'em_bright_create_param_sweep_plot'
    )
    # Define the executable arguments association
    exec_arg_assoc = {
        em_bright_extract_executable:
            " --input $(macroinput) --output $(macrooutput)",
        em_bright_join_executable:
            " --input $(macroinput) --config $(macroconfig)"
            " --output $(macrooutput)",
        em_bright_categorize_executable:
            " --input $(macroinput) --output $(macrooutput)"
            " --eosname $(macroeos)",
        em_bright_train_executable:
            " --input $(macroinput) --config $(macroconfig)"
            " --output $(macrooutput) --param-sweep-plot-prefix $(macroeos)",
        em_bright_paramater_sweep_plot_executable:
            " --input $(macroinput) --config $(macroconfig)"
            " --verbose"
    }
    # create a dictionary of Submit instances
    condor_sub_dict = dict.fromkeys(exec_arg_assoc)
    # create all sub files
    for exect, arg_sub in exec_arg_assoc.items():
        condor_sub_dict[exect] = htcondor.Submit(
            universe='vanilla',
            request_disk="1GB",
            executable=exect,
            arguments=arg_sub,
            output='$(executable).stdout',
            error='$(executable).stderr',
            log='$(executable).log',
            accounting_group=accounting_group
        )
    # Creating list of sqlite files to create the dag
    inj_list = glob(
        os.path.join(
            os.path.abspath(args.file_dir), inj_file_pattern
        )
    )
    # Instantiate the DAG
    condor_dag = dags.DAG()
    # get vars for em_bright_extract layer
    em_bright_extract_vars = []
    for idx, injFilename in enumerate(inj_list):
        match = re.search(sqlite_run_tag, os.path.basename(injFilename))
        if match is None:
            continue
        prefix, startT, duration = match.groups()
        # filename for INJCOINC files
        output_fname = \
            em_bright_extract_prefix + \
            prefix + '-' + startT + '-' + duration + em_bright_extract_suffix
        em_bright_extract_vars.append(
            dict(
                macroinput=injFilename,
                macrooutput=output_fname
            )
        )

    em_bright_extract_layer = condor_dag.layer(
        name=em_bright_extract_executable,
        submit_description=condor_sub_dict[em_bright_extract_executable],
        vars=em_bright_extract_vars
    )
    # output for JOIN
    join_output = \
        em_bright_join_prefix + '-' + startT + '-' + \
        duration + em_bright_join_suffix
    join_output = os.path.join(abs_work_dir, join_output)

    em_bright_join_vars = [
        dict(
            macroinput=abs_work_dir,
            macrooutput=join_output,
            macroconfig=abs_config_file
        )
    ]
    em_bright_join_layer = em_bright_extract_layer.child_layer(
        name=em_bright_join_executable,
        submit_description=condor_sub_dict[em_bright_join_executable],
        vars=em_bright_join_vars
    )

    # input and output for CATEGORIZE/TRAIN
    em_bright_categorize_vars = list()
    em_bright_train_vars = list()
    categorize_input = join_output
    for em_bright_eos in EOS_BAYES_FACTORS:
        categorize_output = \
            em_bright_categorize_prefix + '-' + startT + '-' + duration + \
            '-' + em_bright_eos + em_bright_categorize_suffix
        categorize_output = os.path.join(abs_work_dir, categorize_output)

        train_output = \
            em_bright_train_prefix + '-' + startT + '-' + duration + \
            '-' + em_bright_eos + em_bright_train_suffix
        train_output = os.path.join(abs_work_dir, train_output)

        em_bright_categorize_vars.append(
            dict(
                macroinput=categorize_input,
                macrooutput=categorize_output,
                macroeos=em_bright_eos
            )
        )
        em_bright_train_vars.append(
            dict(
                macroinput=categorize_output,
                macrooutput=train_output,
                macroconfig=abs_config_file,
                macroeos=em_bright_eos
            )
        )
    em_bright_categorize_layer = em_bright_join_layer.child_layer(
        name=em_bright_categorize_executable,
        submit_description=condor_sub_dict[em_bright_categorize_executable],
        vars=em_bright_categorize_vars
    )
    em_bright_train_layer = em_bright_categorize_layer.child_layer(
        name=em_bright_train_executable,
        submit_description=condor_sub_dict[em_bright_train_executable],
        vars=em_bright_train_vars
    )
    em_bright_param_sweep_layer = em_bright_train_layer.child_layer(  # noqa: F841,E501
        name=em_bright_paramater_sweep_plot_executable,
        submit_description=condor_sub_dict[
            em_bright_paramater_sweep_plot_executable],
        vars=[dict(macroinput=abs_work_dir, macroconfig=abs_config_file)]
    )

    print("DAG desription:\n", condor_dag.describe())
    dags.write_dag(condor_dag, abs_work_dir,
                   dag_file_name=args.dagname)

    # FIXME Condor refuses to carry over env vars
    # remove when solution is found
    for exe in config['executables']:
        src = os.path.join(args.executables_dir, exe)
        dst = os.path.join(abs_work_dir, exe)
        if os.path.lexists(dst):
            os.remove(dst)
        shutil.copy(src, dst)
    print('{} is written in {}'.format(args.dagname, abs_work_dir))



import numpy as np
import astropy.units as u
import astropy.constants as c
from scipy.optimize import fsolve
from scipy.interpolate import UnivariateSpline

import lal
import lalsimulation as lalsim

from . import PACKAGE_FILENAMES


def compute_isco(chi_bh):
    '''
    This function takes as input the aligned spin component of the BH and
    returns the innermost stable circular orbit radius normalized by the mass
    of the BH.
    '''
    if not np.all([np.abs(chi_bh) <= 1.0]):
        raise ValueError('|chi1| must be less than 1.0')
    Z1 = 1.0 + ((1.0 - chi_bh**2)**(1./3.))*((1 + chi_bh)**(1./3.) +
                                             (1 - chi_bh)**(1./3.))
    Z2 = np.sqrt(3*chi_bh**2 + Z1**2)
    r_isco = 3 + Z2 - np.sign(chi_bh) * np.sqrt((3 - Z1)*(3 + Z1 + 2*Z2))
    return r_isco


def _lalsim_neutron_star_radius(m, max_mass, fam):
    '''Return neutron star radius in meters'''
    if m > max_mass:
        return m * 1500  # 1 solar mass = 1500m
    else:
        try:
            return lalsim.SimNeutronStarRadius(m * lal.MSUN_SI, fam)
        except RuntimeError:
            # FIXME handle RuntimeError for edge cases
            # Raised if m is close to max_mass
            # GSL function failed: interpolation error (errnum=1)
            # XLAL Error - <GSL function> (interp.c:150): Generic failure
            return m * 1500


def _r_ns_from_lal_simulation(m_ns, eosname):
    eos = lalsim.SimNeutronStarEOSByName(eosname)
    fam = lalsim.CreateSimNeutronStarFamily(eos)
    max_mass = lalsim.SimNeutronStarMaximumMass(fam)/c.M_sun.value
    try:
        iter(m_ns)
        R_ns = np.array([_lalsim_neutron_star_radius(m, max_mass, fam)
                        for m in m_ns])
    except TypeError:
        R_ns = _lalsim_neutron_star_radius(m_ns, max_mass, fam)
    return R_ns, max_mass


def _compactness_baryon_mass(m_ns, r_ns):
    C_ns = c.G * m_ns * c.M_sun/(r_ns * u.m * c.c**2)
    C_ns = C_ns.value

    d1 = 0.619
    d2 = 0.1359
    BE = (d1*C_ns + d2*C_ns*C_ns) * m_ns  # arXiv:1601.06083 (Eq: 21)
    m2_b = m_ns + BE  # Baryonic mass - Gravitational mass = Binding Energy
    return [C_ns, m2_b]


def computeCompactness(M_ns, eosname='2H', R_ns=None,
                       max_mass=2.834648092299807):
    '''
    Return the neutron star compactness as a function of mass
    and equation of state or radius

    Parameters
    ----------
    M_ns : array_like
        Neutron star mass in solar masses
    eosname : str
        Neutron star equation of state to be used
    R_ns : array_like
        Neutron star radius.
    max_mass : float
        Maximum mass of neutron star.

    Returns
    -------
    [C_ns, m2_b, max_mass]
        Compactness, baryon mass and maximum neutron star mass
        in solar masses.

    Notes
    -----
    The radius and maximum mass of the neutron star is
    inferred based on the equation of state supplied.
    Supply explicitly if `eosname=None`.

    Examples
    --------
    >>> computeCompactness(2.8)
    [array(0.298), array(3.354), 2.834]
    >>> computeDiskMass.computeCompactness(2.9, eosname='AP4')
    [0.5, 0.0, 2.212]
    >>> m_ns = np.array([1.1, 1.2, 1.3])
    >>> computeDiskMass.computeCompactness(m_ns, eosname='AP4')
    [array([0.141, 0.154, 0.167]), array([1.199, 1.318, 1.439]), 2.212]
    >>> computeCompactness(2.1, R_ns=16000, eosname=None, max_mass=2.5)
    [0.193, 2.362, 2.5]
    '''
    if R_ns:
        C_ns, m2_b = _compactness_baryon_mass(M_ns, R_ns)
    elif eosname != '2H':
        # infer radius and maximum mass based on lalsimulation EoS
        R_ns, max_mass = _r_ns_from_lal_simulation(M_ns, eosname)
        C_ns, m2_b = _compactness_baryon_mass(M_ns, R_ns)
    else:
        with open(PACKAGE_FILENAMES['equil_2H.dat'], 'rb') as f:
            M_g, M_b, Compactness = np.loadtxt(f, unpack=True)
        max_mass = 2.834648092299807
        s = UnivariateSpline(M_g, Compactness, k=5)
        s_b = UnivariateSpline(M_g, M_b, k=5)
        C_ns = s(M_ns)
        m2_b = s_b(M_ns)
    try:
        C_ns[M_ns > max_mass] = 0.5  # BH compactness set to 0.5
        m2_b[M_ns > max_mass] = 0.0  # BH baryon mass set to 0.0
    except TypeError:  # if C_ns is not an array
        if M_ns > max_mass:
            C_ns = 0.5
            m2_b = 0.0
    return [C_ns, m2_b, max_mass]


'''
For Newtonian: Obtained from arXiv 1807.00011
alpha = 0.406
beta = 0.139
gamma = 0.255
delta = 1.761

For Kerr: Obtained from private message from F. Foucart
alpha = 0.24037857
beta = 0.15184471
gamma = 0.35607169
delta = 1.81491784
'''


def xi_implicit(xi, kappa, chi1, eta):
    xi_implicit = 3*(xi**2 - 2*kappa*xi + kappa**2*chi1**2)
    xi_implicit /= (xi**2 - 3*kappa*xi + 2*chi1*np.sqrt(kappa**3*xi))
    xi_implicit -= eta*xi**3
    return xi_implicit


def computeDiskMass(m1, m2, chi1, chi2, eosname='2H', kerr=False,
                    R_ns=None, max_mass=None):
    """
    This function computes the remnant disk mass after the coalescence using
    the equation (4) arXiv 1807.00011.

    Parameters
    ----------
    m1, m2 : array_like
        primary and secondary mass(es)
    chi1, chi2 : array_like
        primary and secondary spin(s)
    eosname : str
        Name of the equation of state to be used. `AP4`
        `None` when supplying `R_ns`.
    kerr : bool
        Supply to use the relativistic tidal parameter. See Fishbone (1971).
    R_ns : float
        Radius of the secondary, assuming it is a neutron star.
    max_mass : float
        Maximum mass of a neutron star. To be supplied if not supplying EoS.

    Example
    -------
    >>> computeDiskMass(5.0, 2.0, 0.99, 0.)
    0.6321412881595185
    >>> m1 = np.array([5., 6., 7.])
    >>> m2 = np.array([1.0, 1.2, 1.6])
    >>> chi1 = np.zeros(3)
    >>> chi2 = np.zeros(3)
    >>> computeDiskMass(m1, m2, chi1, chi2)
    array([0.12833991, 0.05054819, 0.])
    >>> computeDiskMass(m1, m2, chi1, chi2, eosname='AP4')
    array([0.00851525, 0., 0.])
    >>> max_mass = 3.0  # in m_sun
    >>> r_ns = 15000.  # in meters
    >>> computeDiskMass(m1, m2, chi1, chi2,
    ...                 R_ns=r_ns, max_mass=max_mass, eosname=None)
    array([0.12265712, 0.04272054, 0.])

    Notes
    -----
    The primary mass, by convention, is larger one. If arrays are supplied,
    `m1`, `m2`, `chi1`, `chi1` should be of the same size.
    """
    try:
        _ = iter(m1)
        assert m1.size == m2.size == chi1.size == chi2.size, \
            "masses and spins must have the same dimensions"
        # Making sure that the supplied m1 >= m2 ##
        largerMass1 = m1 >= m2
        largerMass2 = m2 > m1
        m_primary = np.zeros_like(m1)
        m_secondary = np.zeros_like(m2)
        chi_primary = np.zeros_like(chi1)
        chi_secondary = np.zeros_like(chi2)

        m_primary[largerMass1] = m1[largerMass1]
        m_primary[~largerMass1] = m2[largerMass2]
        m_secondary[~largerMass2] = m2[~largerMass2]
        m_secondary[largerMass2] = m1[~largerMass1]

        chi_primary[largerMass1] = chi1[largerMass1]
        chi_primary[~largerMass1] = chi2[~largerMass1]
        chi_secondary[~largerMass2] = chi2[~largerMass2]
        chi_secondary[largerMass2] = chi1[~largerMass1]
    except TypeError:
        if m1 >= m2:
            m_primary = m1
            m_secondary = m2
            chi_primary = chi1
            chi_secondary = chi2
        else:
            m_primary = m2
            m_secondary = m1
            chi_primary = chi2
            chi_secondary = chi1

    [C_ns, m2_b, max_mass] = computeCompactness(m_secondary, eosname=eosname,
                                                R_ns=R_ns, max_mass=max_mass)

    eta = m_primary*m_secondary/(m_primary + m_secondary)**2
    BBH = m_secondary > max_mass
    BNS = m_primary < max_mass
    # FIXME
    if type(BNS) == bool:
        if BNS or BBH:
            return float(not(BBH) or BNS)

    if not kerr:
        alpha = 0.406
        beta = 0.139
        gamma = 0.255
        delta = 1.761
        firstTerm = alpha*(1 - 2*C_ns)/(eta**(1/3))
        secondTerm = beta*compute_isco(chi1)*C_ns/eta
        thirdTerm = gamma
    else:
        alpha = 0.24037857
        beta = 0.15184471
        gamma = 0.35607169
        delta = 1.81491784
        kappa = (m_primary/m_secondary)*C_ns
        try:
            Num = len(kappa)
            xi = fsolve(xi_implicit, 100*np.ones(Num), args=(kappa, chi1, eta))
        except TypeError:
            xi = fsolve(xi_implicit, 100, args=(kappa, chi1, eta))

        firstTerm = alpha*xi*(1 - 2*C_ns)
        secondTerm = beta*compute_isco(chi1)*C_ns/eta
        thirdTerm = gamma
    combinedTerms = firstTerm - secondTerm + thirdTerm

    if hasattr(combinedTerms, 'ndim') and combinedTerms.ndim >= 1:
        lessthanzero = combinedTerms < 0.0
        combinedTerms[lessthanzero] = 0.0
    else:
        combinedTerms = 0 if combinedTerms < 0 else combinedTerms
    M_rem = (combinedTerms**delta)*m2_b
    if type(BNS) == np.ndarray:
        M_rem[BBH] = 0.0  # BBH is always EM-Dark
        M_rem[BNS] = 1.0  # Ad hoc value, assuming BNS is always EM-Bright
    return M_rem



"""Module containing tools for EM-Bright classification of
compact binaries using trained supervised classifier
"""
import h5py

import numpy as np
from scipy.interpolate import interp1d
from astropy import cosmology, units as u

from . import (
    EOS_BAYES_FACTORS,
    PACKAGE_FILENAMES,
    computeDiskMass,
    utils
)


_classifiers = {
    eos: utils._open_and_return_clfs(
        PACKAGE_FILENAMES[f'{eos}.pickle']
    )
    for eos in EOS_BAYES_FACTORS
}
"""Classifiers keyed on EOS names. Order (clf_ns, clf_em)"""
assert set(_classifiers) == set(EOS_BAYES_FACTORS), "Inconsistency in"
" number of trained classifiers."


def mchirp(m1, m2):
    return(m1 * m2)**(3./5.)/(m1 + m2)**(1./5.)


def q(m1, m2):
    return m2/m1 if m2 < m1 else m1/m2


def source_classification(m1, m2, chi1, chi2, snr,
                          ns_classifier=None,
                          emb_classifier=None):
    """
    Computes ``HasNS`` and ``HasRemnant`` probabilities
    from point mass, spin and signal to noise ratio
    estimates.

    Parameters
    ----------
    m1 : float
        primary mass
    m2 : float
        secondary mass
    chi1 : float
        dimensionless primary spin
    chi2 : float
        dimensionless secondary spin
    snr : float
        signal to noise ratio of the signal
    ns_classifier : object, optional
        pickled object for NS classification
    emb_classifier : object, optional
        pickled object for EM brightness classification

    Returns
    -------
    tuple
        (P_NS, P_EMB) predicted values.

    Notes
    -----
    By default the classifiers, trained based
    on different nuclear equations of state (EoSs)
    are downloaded from the project page:
    https://git.ligo.org/deep.chatterjee/em-bright.
    The methodology is described in arXiv:1911.00116.
    The score from each classifier is weighted based on
    the bayes factors of individual EoSs as mentioned in
    Table I of arXiv:2104.08681.
    However, if the trained classifiers are supplied
    via ``ns_classifier`` and ``emb_classifier``,
    the score is reported based on the classifier instead
    of re-weighting the score.
    Examples
    --------
    >>> from ligo.em_bright import em_bright
    >>> em_bright.source_classification(2.0 ,1.0 ,0. ,0. ,10.0)
    (1.0, 1.0)
    """
    features = [[m1, m2, chi1, chi2, snr]]
    try:
        # custom classifiers supplied
        return (
            ns_classifier.predict_proba(features).T[1][0],
            emb_classifier.predict_proba(features).T[1][0]
        )
    except AttributeError as e:
        msg, *_ = e.args
        if msg != """'NoneType' object has no attribute 'predict_proba'""":
            raise

    reweighted_ns_score = reweighted_emb_score = 0.
    for eosname, bayes_factor in EOS_BAYES_FACTORS.items():
        ns_classifier, emb_classifier = _classifiers[eosname]
        reweighted_ns_score += ns_classifier.predict_proba(
            features).T[1][0] * bayes_factor
        reweighted_emb_score += emb_classifier.predict_proba(
            features).T[1][0] * bayes_factor
    return reweighted_ns_score, reweighted_emb_score


def get_redshifts(distances, N=10000):
    """
    Compute redshift using the Planck15 cosmology.

    Parameters
    ----------
    distances: float or numpy.ndarray
              distance(s) in Mpc

    N : int, optional
      Number of steps for the computation of the interpolation function

    Example
    -------
    >>> distances = np.linspace(10, 100, 10)
    >>> em_bright.get_redshifts(distances)
    array([0.00225566, 0.00450357, 0.00674384, 0.00897655,
           0.01120181, 0.0134197 , 0.01563032, 0.01783375
           0.02003009, 0.02221941])

    Notes
    -----
    This function accepts HDF5 posterior samples file and computes
    redshift by interpolating the distance-redshift relation.
    """
    function = cosmology.Planck15.luminosity_distance
    min_dist = np.min(distances)
    max_dist = np.max(distances)
    z_min = cosmology.z_at_value(func=function, fval=min_dist*u.Mpc)
    z_max = cosmology.z_at_value(func=function, fval=max_dist*u.Mpc)
    z_steps = np.linspace(z_min - (0.1*z_min), z_max + (0.1*z_min), N)
    lum_dists = cosmology.Planck15.luminosity_distance(z_steps)
    s = interp1d(lum_dists, z_steps)
    redshifts = s(distances)
    return redshifts


def source_classification_pe(posterior_samples_file, hdf5=True,
                             threshold=3.0, sourceframe=True):
    """
    Compute ``HasNS`` and ``HasRemnant`` probabilities from posterior
    samples.

    Parameters
    ----------
    posterior_samples_file : str
        Posterior samples file

    hdf5 : bool, optional
        Supply when not using HDF5 format

    threshold : float, optional
        Maximum neutron star mass for `HasNS` computation

    sourceframe : bool, optional
        Supply to use detector frame quantities

    Returns
    -------
    tuple
        (P_NS, P_EMB) predicted values.


    Examples
    --------
    >>> from ligo.em_bright import em_bright
    >>> em_bright.source_classification_pe('posterior_samples.hdf5')
    (1.0, 0.9616727412238634)
    >>> em_bright.source_classification_pe('posterior_samples.dat', hdf5=False)  # noqa:E501
    (0.0, 0.0)
    """
    if hdf5:
        with h5py.File(posterior_samples_file, 'r') as data:
            engine = list(data['lalinference'].keys())[0]
            samples = data['lalinference'][engine]['posterior_samples'][()]
        mc_det_frame = samples['mc']
        lum_dist = samples['dist']
        redshifts = get_redshifts(lum_dist)
        if sourceframe:
            mc = mc_det_frame/(1 + redshifts)
        else:
            mc = mc_det_frame

    else:
        samples = np.recfromtxt(posterior_samples_file, names=True)
        if sourceframe:
            mc = samples['mc_source']
        else:
            mc = samples['mc']

    q = samples['q']
    m1 = mc * (1 + q)**(1/5) * (q)**(-3/5)
    m2 = mc * (1 + q)**(1/5) * (q)**(2/5)

    try:
        chi1 = samples['a1'] * np.cos(samples['tilt1'])
        chi2 = samples['a2'] * np.cos(samples['tilt2'])
    except ValueError:
        # for aligned-spin PE, a1, a2 is the z component
        chi1 = samples['a1']
        chi2 = samples['a2']

    M_rem = computeDiskMass.computeDiskMass(m1, m2, chi1, chi2)
    prediction_ns = np.sum(m2 <= threshold)/len(m2)
    prediction_em = np.sum(M_rem > 0)/len(M_rem)

    return prediction_ns, prediction_em



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ligo', 'ligo.em_bright', 'ligo.em_bright.tests']

package_data = \
{'': ['*'], 'ligo.em_bright.tests': ['data/*']}

install_requires = \
['astropy>=5.1',
 'h5py>=2.10.0',
 'lalsuite>=7.0,<8.0',
 'pandas>=1.0,<2.0',
 'scikit-learn>=1.1,<2.0']

entry_points = \
{'console_scripts': ['em_bright_categorize = ligo.em_bright.categorize:main',
                     'em_bright_categorize_all_eos = '
                     'ligo.em_bright.categorize:main_all',
                     'em_bright_create_param_sweep_plot = '
                     'ligo.em_bright.utils:param_sweep_plot',
                     'em_bright_dag_writer = ligo.em_bright.dag_writer:main',
                     'em_bright_extract = ligo.em_bright.utils:extract',
                     'em_bright_join = ligo.em_bright.utils:join',
                     'em_bright_train = ligo.em_bright.utils:train']}

setup_kwargs = {
    'name': 'ligo.em-bright',
    'version': '1.0.2',
    'description': 'Possibility and properties of Electromagnetically-bright sources of gravitational-wave events',
    'long_description': None,
    'author': 'Deep Chatterjee',
    'author_email': 'deep.chatterjee@ligo.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
